
signature TRANSFER =
sig
  
  val transfer_tycon : (theory * string) -> (string * mixfix option) Symtab.table -> 
                       (string -> string) -> theory -> 
                       (string * int * (string * string) option * theory)
  val transfer_const : bool -> (theory * string) -> (Term.typ -> Term.typ) -> 
                       (string * mixfix option) Symtab.table -> (string -> string) -> theory -> 
                       (string * theory)
  val transfer_def   : bool -> (theory * string) -> string -> (Term.term -> Term.term) -> 
                       (string * mixfix option) Symtab.table -> (string -> string) -> theory -> 
                       (string * theory)
  val transfer_axiom : (theory * string) -> (Term.term -> Term.term) -> 
                       (string * mixfix option) Symtab.table -> (string -> string) -> theory -> 
                       (string * theory) 
  val transfer_syntax : (Term.typ -> Term.typ) -> string list -> theory -> theory -> theory

  val mk_arity_type_id : string -> string

end

structure Transfer : TRANSFER =
struct

fun mk_arity_type_id type_id = (case (Aux.unpack type_id) of
                                         [q, nm] => (q ^ ".arity_type_" ^ nm)
                                       | [q, q', nm] => (q ^ "." ^ q' ^ ".arity_type_" ^ nm)
                                       | _ => error ("Unexpected path in type constructor name: "^type_id^
                                                     " in mk_arity_type_id"))


fun transfer_tycon (src, tc) renames find_name trg =
 let val trg_id = Aux.name_of_thy trg
     val (tc', mxfx) = (case (Symtab.lookup renames tc) of
                         SOME (r, mx) => (let val name = (trg_id ^ "." ^ r)
                                          in
                                          (case mx of
                                            SOME m => (let val _ = Syn_mng.add_type_mixfix (name, m)
                                                       in
                                                        (name, m)
                                                       end)
                                          | NONE => (case (Syn_mng.get_type_mixfix tc) of
                                                       NONE => (name, NoSyn)
                                                     | SOME m => 
                                                       (let val _ = Syn_mng.add_type_mixfix (name, m)
                                                        in
                                                         (name, m)
                                                        end)))
                                          end)
                       | NONE => (let val (q, nm) = Aux.qname_split tc
                                      val b_nm = find_name nm
                                      val tc' = (trg_id ^ "." ^ b_nm)
                                      val mxfx = (case (Syn_mng.get_type_mixfix tc) of
                                                   NONE => NoSyn
                                                 | SOME m => (let val _ = Syn_mng.add_type_mixfix (tc', m)
                                                              in
                                                                m
                                                              end))
                                  in
                                  (tc', mxfx)
                                  end))  
  in
    case (Aux.type_arity(src, tc)) of
     SOME (arity, i) => 
       (let val arities = map (fn (cls, srtl) => (tc', flat srtl, cls)) arity
            val _ = Aux.info ("... adding type constructor " ^ tc')
            val trg' = trg |> Theory.add_types [(NameSpace.base tc', i, mxfx)] 
                           |> (fold AxClass.axiomatize_arity_cmd) arities                   
        in
         (tc', i, SOME (mk_arity_type_id tc, mk_arity_type_id tc'), trg')
        end)
   | NONE => (let val _ = Aux.info ("... adding non-logical type constructor " ^ tc')
                  val trg' = trg |> Theory.add_types [(NameSpace.base tc', 0, mxfx)]
              in
               (tc', 0, NONE, trg')
              end) 
  end



fun transfer_const logical (src, opnm) map_type renames find_name trg =
  let val ty = case (Aux.const_type src opnm) of
                NONE => error ("Unexpected error in transfer_const") 
              | SOME t => Aux.unvarifyT t
      val trg_id = Aux.name_of_thy trg
      fun string_of_typ thy = (Pretty.string_of o (Sign.pretty_typ thy))  

      val (((path, opnm_b), opnm_f), mxfx) = 
                         (case (Symtab.lookup renames opnm) of
                            SOME (r, mx) => (let val (p, b) = Aux.qname_split r
                                                 val name = if p = "" then ((p, b), trg_id ^ "." ^ b)
                                                            else ((p, b), trg_id ^ "." ^ p ^"."^ b)
                                             in
                                              (case mx of
                                                SOME m => (let val _ = Syn_mng.add_mixfix (snd name, m)
                                                           in
                                                            (name, m)
                                                           end)
                                              | NONE => (case (Syn_mng.get_mixfix opnm) of
                                                          NONE => (name, NoSyn)
                                                        | SOME m =>(let val _ = Syn_mng.add_mixfix (snd name, m)
                                                                    in
                                                                      (name, m)
                                                                    end)))
                                              end)
                           | NONE => (let val (q, nm) = Aux.qname_split2 opnm
                                          val b_opnm = find_name nm
                                          val (p, b) = Aux.qname_split b_opnm
                                          val opnm' = if p = "" then ((p, b), trg_id ^ "." ^ b)
                                                      else ((p, b), trg_id ^ "." ^ p ^"."^ b)
                                          val mxfx = (case (Syn_mng.get_mixfix opnm) of
                                                       NONE => NoSyn
                                                     | SOME m => (let val _ = Syn_mng.add_mixfix (snd opnm', m)
                                                                  in
                                                                   m
                                                                  end))
                                      in
                                        (opnm', mxfx)
                                      end))

        val ty' = map_type ty
        val _ = if logical then Aux.info ("... adding logical constant \n" ^ opnm_f ^ " :: \"" ^ 
                               (string_of_typ trg ty') ^ "\"   " ^ (Syn_mng.string_of_mixfix mxfx))
                else Aux.info ("... adding derived constant \n" ^ opnm_f ^ " :: \"" ^ 
                               (string_of_typ trg ty') ^ "\"   " ^ (Syn_mng.string_of_mixfix mxfx))

        val trg' = (if (path = "") then trg |> Theory.add_consts_i [(opnm_b, ty', mxfx)]
                    else trg
                         |> Sign.add_path path
                         |> Theory.add_consts_i [(opnm_b, ty', mxfx)]
                         |> Sign.parent_path)
  in 
   (opnm_f, trg')
  end


fun transfer_def overloaded (src, defnm) opnm map_term renames find_name trg =
 let val ax = get_axiom src defnm
     val term = (Aux.unvarify o prop_of) ax
     val trg_id = Aux.name_of_thy trg                                     
     val ((path, defnm_b), defnm_f) = 
             (case (Symtab.lookup renames defnm) of
                    SOME (r, _) => let val (p, b) = Aux.qname_split r
                                   in
                                    if p = "" then ((p, b), trg_id ^ "." ^ b)
                                    else ((p, b), trg_id ^ "." ^ p ^"."^ b)
                                   end
                  | NONE => (let val (dq, dnm) = Aux.qname_split2 defnm
                                 val b_defnm = find_name dnm
                                 val (p, b) = Aux.qname_split b_defnm
                             in 
                               if p = "" then ((p, b), trg_id ^ "." ^ b)
                               else ((p, b), trg_id ^ "." ^ p ^"."^ b)
                             end))
     val _ = (if overloaded 
              then Aux.info ("... adding definition (overloaded) " ^ defnm_f 
                                ^ " of the constant " ^ opnm)
              else Aux.info ("... adding definition " ^ defnm_f
                                ^ " of the constant " ^ opnm))

     val tterm = map_term term
     val (gen_thms, trg') = 
     (if (path = "") then trg |> PureThy.add_defs_i overloaded [((defnm_b, tterm), [])]
      else trg 
           |> Sign.add_path path
           |> PureThy.add_defs_i overloaded [((defnm_b, tterm), [])]
           ||> Sign.parent_path) 
 in
  (defnm_f, trg')
 end

fun transfer_axiom (src, axnm) map_term renames find_name trg =
 let val ax = get_axiom src axnm
     val term = (Aux.unvarify o prop_of) ax
     val trg_id = Aux.name_of_thy trg
     val ((path, axnm_b), axnm_f) = 
             (case (Symtab.lookup renames axnm) of
                    SOME (r, _) => (let val (p, b) = Aux.qname_split r
                                    in
                                     if p = "" then ((p, b), trg_id ^ "." ^ b)
                                     else ((p, b), trg_id ^ "." ^ p ^"."^ b)
                                    end)
                  | NONE => (let val (aq, anm) = Aux.qname_split2 axnm
                                 val b_axnm = find_name anm
                                 val (p, b) = Aux.qname_split b_axnm
                             in 
                               if p = "" then ((p, b), trg_id ^ "." ^ b)
                               else ((p, b), trg_id ^ "." ^ p ^"."^ b)
                             end))                          
     val _ = Aux.info ("... translating axiom " ^ axnm ^ " as " ^ axnm_f ^ " ...")
     val tterm = map_term term
     val tterm_rep = Sign.string_of_term trg tterm
     val (gen_thms, trg') = 
     (if (path = "") then trg |> PureThy.add_axioms_i [((axnm_b, tterm), [])]
      else trg 
           |> Sign.add_path path
           |> PureThy.add_axioms_i [((axnm_b, tterm), [])]
           ||> Sign.parent_path) 
     val _ = Aux.warning ("... added axiom " ^ axnm_f ^ "\n" ^ quote(tterm_rep))          
 in
  (axnm_f, trg')
 end




(* *** Syntax *** *)


fun transfer_syntax map_type domain source target =
  let val trgnm = Aux.name_of_thy target

      fun nts u = Aux.foldl (op @) ([], map Syn_mng.get_nonterms domain)
      fun syn_decls u = Aux.foldl (op @) ([], map Syn_mng.get_syn_decls domain)
      fun trrules u = Aux.foldl (op @) ([], map Syn_mng.get_trrules domain)
      fun parse_ast_trans u = Aux.foldl (op @) ([], map Syn_mng.get_parse_ast_trans domain)
      fun parse_trans u = Aux.foldl (op @) ([], map Syn_mng.get_parse_trans domain)
      fun print_trans u = Aux.foldl (op @) ([], map Syn_mng.get_print_trans domain)
      fun print_ast_trans u = Aux.foldl (op @) ([], map Syn_mng.get_print_ast_trans domain)
      fun token_trans u = Aux.foldl (op @) ([], map Syn_mng.get_token_trans domain)



      fun ins_nt (thy, nt) = (if (Syn_mng.is_transl1 nt thy) then thy
                              else (let val _ = Aux.info ("... adding nonterminal: " ^ nt)
                                        val _ = Syn_mng.add_transl1 nt trgnm
                                        val _ = Syn_mng.add_nonterm (trgnm, nt)
                                    in 
                                     Sign.add_nonterminals [nt] thy
                                    end))
      val trg2 = Aux.foldl ins_nt (target, nts ())

      fun transl_tp (Type(nm, ts)) = 
          let val unm = Aux.unpack nm
          in
           if ((hd unm) mem domain) then Type (Aux.pack (trgnm::(tl unm)), map transl_tp ts)
           else Type (nm, map transl_tp ts)    
          end
        | transl_tp var = var 

      fun transl_syn_decl (thy, (s, b, nm, tp, mxfx)) =
       (if (Syn_mng.is_transl1 nm thy) then thy
        else (let val ntp = Aux.unvarifyT (transl_tp (map_type tp))
                  val _ = Aux.info ("... adding syntax declaration: " ^ nm ^ 
                                       " :: \"" ^ (Sign.string_of_typ thy ntp) ^ 
                                       "\"   " ^ (Syn_mng.string_of_mixfix mxfx))
                  val _ = Syn_mng.add_transl1 nm trgnm
                  val _ = Syn_mng.add_syn_decl (trgnm, (s, b, nm, ntp, mxfx))
              in 
               Sign.add_modesyntax_i (s, b) [(nm, ntp, mxfx)] thy
              end))

      val trg3 = Aux.foldl transl_syn_decl (trg2, syn_decls ()) 


(* ***** *)

      fun show_trrule (Syntax.ParseRule (l, r)) = ("\"" ^ snd l ^ "\"" ^ " => " ^ "\"" ^ snd r ^ "\"" )
        | show_trrule (Syntax.PrintRule (l, r)) = ("\"" ^ snd l ^ "\"" ^ " <= " ^ "\"" ^ snd r ^ "\"")
        | show_trrule (Syntax.ParsePrintRule (l, r)) = ("\"" ^ snd l ^ "\"" ^ " == " ^ "\"" ^ snd r ^ "\"")


      fun ins_trrule (thy, trrule) = 
       let val rep = show_trrule trrule
       in
        if (Syn_mng.is_transl1 rep thy) then thy
        else (let val _ = Aux.info ("... adding translation: " ^ rep)
                  val _ = Syn_mng.add_trrule (trgnm, trrule)
                  val _ = Syn_mng.add_transl1 rep trgnm
              in
                Sign.add_trrules [trrule] thy
              end)
       end
      val trg4 = Aux.foldl ins_trrule (trg3, trrules ())
 
      fun ins_parse_ast (thy, t) = 
       let val rep = snd t
       in
        if (Syn_mng.is_transl1 rep thy) then thy
        else (let val _ = Aux.info ("... adding parse_ast_translation: " ^ rep)
                  val _ = Syn_mng.add_parse_ast_trans (trgnm, t)
                  val _ = Syn_mng.add_transl1 rep trgnm
              in
               IsarCmd.parse_ast_translation t thy
              end)
       end
      val trg5 = Aux.foldl ins_parse_ast (trg4, parse_ast_trans ())
     
      fun ins_parse (thy, t) = 
       let val rep = snd t
       in
        if (Syn_mng.is_transl1 rep thy) then thy
        else (let val _ = Aux.info ("... adding parse_translation: " ^ rep)
                  val _ = Syn_mng.add_parse_trans (trgnm, t)
                  val _ = Syn_mng.add_transl1 rep trgnm
              in
               IsarCmd.parse_translation t thy
              end)
       end
      val trg6 = Aux.foldl ins_parse (trg5, parse_trans ())

      fun ins_print (thy, t) = 
       let val rep = snd t
       in
        if (Syn_mng.is_transl1 rep thy) then thy
        else (let val _ = Aux.info ("... adding print_translation: " ^ rep)
                  val _ = Syn_mng.add_print_trans (trgnm, t)
                  val _ = Syn_mng.add_transl1 rep trgnm
              in
               IsarCmd.print_translation t thy
              end)
       end
      val trg7 = Aux.foldl ins_print (trg6, print_trans ())
   
      fun ins_print_ast (thy, t) = 
       let val rep = snd t
       in
        if (Syn_mng.is_transl1 rep thy) then thy
        else (let val _ = Aux.info ("... adding print_ast_translation: " ^ rep)
                  val _ = Syn_mng.add_print_ast_trans (trgnm, t)
                  val _ = Syn_mng.add_transl1 rep trgnm
              in
               IsarCmd.print_ast_translation t thy
              end)
       end
      val trg8 = Aux.foldl ins_print_ast (trg7, print_ast_trans ())

      fun ins_token (thy, t) = 
        if (Syn_mng.is_transl1 t thy) then thy
        else (let val _ = Aux.info ("... adding token_translation: " ^ t)
                  val _ = Syn_mng.add_token_trans (Aux.name_of_thy thy, t)
                  val _ = Syn_mng.add_transl1 t trgnm
              in
               IsarCmd.token_translation t thy
              end)

      val trg9 = Aux.foldl ins_token (trg8, token_trans ())
 in
  trg9
 end

end